"""MongoDB DAO Package

Module Poem
-----------

I came up once and hollered!
I came up twice and cried!
If that water hadn't a-been so cold
I might've sunk and died.

But it was Cold in that water! It was cold!

- Langston Hughes (Life is fine)
"""


from geocsvlib.dao.mongodb.ip_geolocation import IPGeolocation



from unittest import TestCase

from geocsvlib.constants import CONFIG_DEFAULT
from geocsvlib.parsers import loader

class LoaderTest(TestCase):
    """Load Configuration Test
    """

    def setUp(self):
        """Set up configuration
        """
        self.response = loader.parse_json_file(CONFIG_DEFAULT)
    
    def tearDown(self):
        """Teardown expensive ops
        """

    def test_json_loading(self):
        """Tests that json data loading function works as expected
        """
        self.assertIsNotNone(self.response)
        self.assertIsInstance(self.response, dict)
    
    def test_json_loading_correctly(self):
        """
        Ensure that the json loaded and retrieved is correct
        """
        #: remove to constants later
        default = {
            "vendor": "mongodb",
            "database": "geocsvlib",
            "port": 27017,
            "url": "localhost",
            "username": "optional",
            "password": "optional",
            "csv_delimiter": ","
        }
        self.assertListEqual(self.response.keys(), default.keys())



"""SQLite DAO Package

Module Poem
-----------

Oh whence do you come, my dear friend, to me,
With your golden hair all fallen below your knee,
And your face as white as snowdrops on the lea,
And your voice as hollow as the hollow sea?

- Christina Rosseti (The Poor Ghost)
"""


from geocsvlib.dao.sqlite.ip_geolocation import IPGeolocation



"""
Application wide logger, could have used module based but making it trivial to re-implement



Logger module uses static programming language type syntax i.e. camelCase,
don't know why they decided on that
but to keep the variables consistent naming convention follows
the camelCase antecedent

Logging will be stored in a log server eventually where all logs
can be viewed as required, this is why
the name of the application is important
"""
import os
import logging

#: Get the environmental configuration or currently configured config
currentConfig = "Development"

#: Necessary to differentiate these logs from logs from other micro services
#: especially when used by the service monitor to trace request path
applicationName = "Parsito"
logLevel = logging.INFO


logger = logging.getLogger(applicationName)
logger.setLevel(logLevel)

#: This can be redirected elsewhere for log recording
loggerHandler = logging.FileHandler(filename='application.log')
loggerHandler.setLevel(logLevel)

format_schema = '%(asctime)s -:::- %(name)s -:::- %(levelname)s -:::- %(message)s'
loggerFormatter = logging.Formatter(format_schema)
loggerHandler.setFormatter(loggerFormatter)


logger.addHandler(loggerHandler)


def message_logger(message, cause, reason=None):
    """
    Formats the message to a consistent format by logging the error and the causing client

    :param message:     The message to be formatted and worked upon
                        :type <type, 'str'>

    :param cause:       The cause of the message being logged, it can for instance be The ip address
                        of the client or transaction ID or operation description.
                        :type <type, 'str'>

    :param reason:      Short string prepositions that tie the relationship of the cause to the message
                        could be limited to any of the following values ['by', 'when', 'in']
                        :type <type, 'str'>

    :return message:    Properly formatted message consistent with the type necessary to provide best
                        information when inspected from a log file or log server whichever is used
                        :type <type, 'str'>
    """
    if not reason:
        reason = 'Error BY'

    message = '{0} : {2} {1}'.format(message, cause, reason)
    return message



"""Model Factory Python Module

Model factory module returns Model DAOs that correspond
to the options provided in the config file


Module Poem
-----------

I went down to the river,
I set down on the bank.
I tried to think but couldn't,
So I jumped in and sank.

- Langston Hughes (Life is fine)
"""
from geocsvlib import get_configuration_value
from geocsvlib.constants import MONGODB, SQLITE, POSTGRES

from geocsvlib.dao.mongodb.ip_geolocation import IPGeolocation
from geocsvlib.dao.sqlite import IPGeolocation as IPGeolocationRecord


class ModelFactory(object):
    """
    ModelFactory returns an appropriate model to use for persistence
    depending on the value provided in the library's config.json
    file
    """
    @staticmethod
    def find(**kwargs):
        """
        Virtual model abstraction that encapsulates the different db backends
        to provide a single interface to load data from the library
        """
        try:
            vendor = get_configuration_value("vendor")
        except KeyError:
            print("You need to use load geocsvlib with a configurator before use")
            return False, {}

        if vendor == MONGODB:
            return IPGeolocationDocument.finder(**kwargs)
        if vendor == SQLITE:
            # return IPGeolocationRecord
            raise NotImplementedError("sqlite support pending")
        if vendor == POSTGRES:
            # NOT YET IMPLEMENTED PROBABLY A GREAT IDEA TO RAISE AN EXC
            raise NotImplementedError("postgres support pending")

    @staticmethod
    def get_model():
        """
        Returns appropriate model to use for persistence
        """
        vendor = get_configuration_value("vendor")
        if vendor == MONGODB:
            return IPGeolocationDocument
        if vendor == SQLITE:
            # return IPGeolocationRecord
            raise NotImplementedError("sqlite support pending")
        if vendor == POSTGRES:
            # NOT YET IMPLEMENTED PROBABLY A GREAT IDEA TO RAISE AN EXC
            raise NotImplementedError("postgres support pending")



"""
If a constant name changes or is removed; Go to war
"""
from unittest import TestCase


class ConstantsTest(TestCase):
    """Constants TestCase class

    Trivial class at the moment and as such no setups or teardowns necessary
    """
    def test_flag_imports(self):
        """Test constants are importable
        A simple import is enough to test the presence of wanted constants
        """
        from geocsvlib.constants import CONFIG_DEFAULT, CONFIG_FLAG, CONFIG_FLAG_HELP
        from geocsvlib.constants import CSV_HELP, DESCRIPTION



"""
Connection Factory Python Module

Connection factory returns database connection loaders based on
the option provided in the config file

Module Poem
-----------

I stand amid the roar
Of a surf-tormented shore,
And I hold within my hand
Grains of the golden sand--
How few! yet how they creep
Through my fingers to the deep,
While I weep--while I weep!
O God! can I not grasp
Them with a tighter clasp?
O God! can I not save
One from the pitiless wave?
Is all that we see or seem
But a dream within a dream?

-Edgar Allan Poe (A Dream Within A Dream)
"""
from mongoengine import connect


class ConnectionFactory(object):
    """
    Returns appropriate connection to db
    """
    @staticmethod
    def connect(configuration):
        """
        Uses default connection or connection from options for
        code based configuration
        """
        import geocsvlib
        from geocsvlib.constants import MONGODB, SQLITE, POSTGRES
        database = configuration.get("database", "geocsvlib")
        vendor = configuration.get("vendor", "mongodb")
        port = configuration.get("port", 27017)
        host = configuration.get("url")
        config = {
            "vendor": vendor,
            "database": database,
            "port": port,
            "url": host,
            "username": "",
            "password": "",
            "csv_delimiter": ","
        }
        if vendor == MONGODB:
            geocsvlib.init(config)
            return connect(database, host=host, port=port)
        if vendor == SQLITE:
            raise NotImplementedError("sqlite support pending")
        if vendor == POSTGRES:
            raise NotImplementedError("postgres support pending")
        else:
            raise Exception("Invalid Library Configuration")

    @staticmethod
    def get_connection():
        """
        Get a connection to the provided database name and as provided
        in the configuration file and also on the config database vendor type
        """
        import geocsvlib
        from geocsvlib.constants import MONGODB, SQLITE, POSTGRES
        # use the vetted config to get the database
        vendor = geocsvlib.get_configuration_value("vendor")
        database = geocsvlib.get_configuration_value("database")
        if vendor == MONGODB:
            return connect(database)
        if vendor == SQLITE:
            raise NotImplementedError("sqlite support pending")
        if vendor == POSTGRES:
            raise NotImplementedError("postgres support pending")
        else:
            raise Exception("Invalid Library Configuration")



"""Reader Python Module

Parses CSV file and persists

Module Poem
-----------
I am not yours, not lost in you,
Not lost, although I long to be
Lost as a candle lit at noon,
Lost as a snowflake in the sea

- Sara Teasdale (I am not yours)
"""
import csv
import numbers
import sys

from mongoengine import connect
from mongoengine import NotUniqueError
from mongoengine import ValidationError

from geocsvlib.utils import logger

from geocsvlib.models import ModelFactory
from geocsvlib.models import ConnectionFactory


class Reader(object):
    """Reader CSV Extraction Class

    parse csv file
    """
    def __init__(self, csv_filepath):
        self.csv_location = csv_filepath

        self.passed = 0
        self.failed = 0
        self.records_processed = 0
        self.total_time_taken = 0

        self.Model = ModelFactory.get_model()
        self.Connection = ConnectionFactory.get_connection()

    def _has_prerequisites(self, ip_address):
        """Minimally alidates semantic correctness of provided data"""
        subnets = ip_address.split(".")
        return len(subnets) == 4 and all(0 <= int(subnet) < 256 for subnet in subnets)

    def _is_numbery(self, value):
        """
        Validate that a value is an instance of a number
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    def _is_processable(self, datum):
        """Parse CSV

        Parse CSV datum where datum is a list of values
        """
        if self._has_prerequisites(datum.get("ip_address")) and self._is_numbery(datum.get("mystery_value")):
            return True
        return False

    def log_csv_stats(self, outcome):
        """
        Log the statistics to be displayed at end of program run
        """
        self.records_processed += 1
        if outcome:
            self.passed += 1
        else:
            self.failed += 1

    def show_csv_stats(self, start_time, end_time):
        """
        Shows stats to stdout on the execution time and details of the CSV parsing
        """
        self.total_time_taken = end_time - start_time

        print "\n"
        print "CSV Records Processed: %r" % self.records_processed
        print "Total Time Taken (seconds): %r" % self.total_time_taken
        print "Successfully Processed and Saved: %r" % self.passed
        print "Failed to Saved: %r" % self.failed

    def process_csv_data(self):
        """Process CSV Data

        This will be much faster if we can guarantee multiprocessor environments
        """
        try:
            with open(self.csv_location, 'rb') as filestream:
                contents = csv.DictReader(filestream)
                for content in contents:
                    # always reset outcome status on new csv record
                    sys.stdout.write("\rSaving record {0} of {1}".format(contents.line_num, "Many"))
                    sys.stdout.flush()
                    status = False

                    if self._is_processable(content):
                        geocsv = self.Model(**content)
                        try:
                            status = geocsv.save()
                        except ValidationError as e:
                            msg = "Could not save GeoCSVLocation"
                            logger.message_logger(msg, e.message)
                        except NotUniqueError as e:
                            msg = "A Similar Resource Already saved On"
                            logger.message_logger(msg, e.message)
                        finally:
                            self.log_csv_stats(status)
                    else:
                        self.log_csv_stats(status)
        except IOError:
            msg = "{0} could not be opened. Check for sufficient privileges or file exists"
            print msg.format(self.csv_location)



"""Loader Python Module

Load provided json configuration file for database connectivity
"""
import json


def parse_json_file(json_file_url):
    """
    Parse provided json file and save as global configuration object for
    later library reuse
    """
    with open(json_file_url) as json_configuration:
        try:
            configuration = json.load(json_configuration)
            return configuration
        except AttributeError:
            print("Invalid json file detected, please try again later")
        except ValueError:
            print("Invalid data file, could not decode into json")



"""Constants Utility Package

Holds library wide constants to provide a single point of concern/editing
for application/library constants

Module Poem
-----------

here is the deepest secret nobody knows
(here is the root of the root and the bud of the bud
and the sky of the sky of a tree called life;which grows
higher than the soul can hope or mind can hide)
and this is the wonder that's keeping the stars apart

I carry your heart(I carry it in my heart)

- EE Cummings (I Carry Your Heart With Me)
"""
CONFIG_DEFAULT = 'config.json'
CONFIG_FLAG = '--config'
CONFIG_FLAG_HELP = 'config.json file to use for database configuration'

CSV_HELP = 'Name of the CSV file to be parsed'

DESCRIPTION = 'Geolocation CSV Parser, Persistor, and Loader Library'

MONGODB = "mongodb"
SQLITE = "sqlite"
POSTGRES = "postgres"
ALLOWED_DATABASES = [MONGODB, SQLITE, POSTGRES]


__all__ = [
    'MONGODB'
]



from unittest import TestCase

from geocsvlib.models import Model


class ConnectionFactoryTest(TestCase):
    """
    Test Connection Factory actions
    """
    def setUp(self):
        pass

    def tearDown(self):
        pass



"""Data Access Objects Package


Poem of the Module
------------------

A Dream Within A Dream

Take this kiss upon the brow!
And, in parting from you now,
Thus much let me avow--
You are not wrong, who deem
That my days have been a dream;
Yet if hope has flown away
In a night, or in a day,
In a vision, or in none,
Is it therefore the less gone?
All that we see or seem
Is but a dream within a dream

- Edgar Allan Poe
"""



"""IPGeolocation Data Access Object

MongoDB variant of the DAO for bidirectional (ingress and egress)
flow of data.

Module Poem
-----------

The free bird leaps
on the back of the wind
and floats downstream
till the current ends
and dips his wings
in the orange sun rays
and dares to claim the sky

- Maya Angelou
"""
from mongoengine import Document
from mongoengine import FloatField
from mongoengine import IntField
from mongoengine import StringField



class IPGeolocation(Document):
    """GeoCSV
    Document"""
    ip_address = StringField(required=True, unique=True)
    city = StringField(required=True)
    country = StringField(required=True)
    country_code = StringField(required=True)
    longitude = FloatField(required=True)
    latitude = FloatField(required=True)
    mystery_value = IntField(required=True)

    @staticmethod
    def finder(**kwargs):
        try:
            result = IPGeolocation.objects(**kwargs).first()
        except:
            # log this exception
            return False, {}
        return (True, result) if result else (False, {})



"""IPGeolocation Data Access Object

SQLite variant of the DAO for bidirectional (ingress and egress)
flow of data.

Module Poem
-----------

Here they be drunken pirates who can not swim
Drunk on ale, Drunk on rum, Drunk on waves of sand

- Raikishi (The War Song)
"""


class IPGeolocation(object):
    """
    IPGeolocation sqlite model interface
    """
    pass



"""GeoCSVlib

Parses Geolocation data from CSV files for persistence and lookup

Module Poem
-----------
Two roads diverged in a yellow wood,
And sorry I could not travel both
And be one traveler, long I stood
And looked down one as far as I could
To where it bent in the undergrowth;
Then took the other, as just as fair,
And having perhaps the better claim,
Because it was grassy and wanted wear;
Though as for that the passing there
Had worn them really about the same,
And both that morning equally lay
In leaves no step had trodden black.
Oh, I kept the first for another day!
Yet knowing how way leads on to way,
I doubted if I should ever come back.
I shall be telling this with a sigh
Somewhere ages and ages hence:
Two roads diverged in a wood, and I-
I took the one less traveled by,
And that has made all the difference

- Robert Frost (The Road Not Taken)
"""


_configuration = {"config": {
    "vendor": "mongodb",
    "database": "pyishdb"
}}


def init(configuration):
    """Initialize Configuration of Library"""
    _configuration["config"] = configuration


def get_configuration_value(keyword):
    """
    Returns the value of a configuration key if set
    """
    return _configuration["config"].get(keyword)



#!/usr/bin/env python
import codecs

from setuptools import setup
from setuptools import find_packages


VERSION = (0, 1, 6)
SEPARATOR = '.'


def get_version():
    return SEPARATOR.join(map(lambda x: str(x), VERSION))

__version__ = get_version()


def parse(*args, **kwargs):
    """
    Parse documentation files into internal variable for rendering with setup
    """
    # noinspection PyPep8Naming
    SEPARATOR = kwargs.get('separator', '\n')
    coding = kwargs.get('coding', 'UTF-8')
    stream = []

    for arg in args:
        with codecs.open(arg, 'r', coding) as document:
            stream.append(document.read())

    return SEPARATOR.join(stream)


#: LONG_DESCRIPTION = parse('README.rst')

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Utilities",
    "Topic :: Software Development"
]


setup(
    name='geocsvlib',
    description='IP Geolocation data CSV Parser Library',
    url='http://github.com/rayattack/geocsvlib',
    version=__version__,
    license='MIT',
    author='Raymond Ortserga',
    author_email='ortserga@yahoo.com',
    zip_safe=False,
    packages=find_packages(),
    install_requires=['mongoengine'],
    keywords='IP Location Parser CSV',
    classifiers=CLASSIFIERS,
    long_description="",
    test_suite='nose.collector',
    tests_require=['nose'],
    scripts=['geocsvlib/geocsv']
)



"""Parser Package

Responsible for connecting CSV parsing functionality to persistence

Module Poem
-----------
But why did your tears soak thro' the clay,
And why did your sobs wake me where I lay?
I was away, far enough away:
Let me sleep now till the Judgment Day

- Christina Rosseti (The Poor Ghost)
"""

from geocsvlib.reader import Reader


__all__ = [
    'Reader'
]



"""Models Package

Holds Data Access Factories and Binds Factories to DAOs

Module Poem
-----------
What happens to a dream deferred?

Does it dry up
Like a raisin in the sun?

-Langston Hughes (Dream Deffered)
"""

from .connection_factory import ConnectionFactory
from .model_factory import ModelFactory

Model = ModelFactory.get_model()

__all__ = [
    'ConnectionFactory',
    'ModelFactory'
]
